# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np
import pandas as pd

from pylife.utils import histogram as hi


@pytest.fixture
def empty_histogram():
    return pd.Series(dtype=np.float64, index=pd.IntervalIndex.from_tuples([]))


@pytest.fixture
def some_histogram_1d():
    return pd.Series(np.array([5., 10.]), index=pd.interval_range(start=0, end=2))


@pytest.fixture
def another_histogram_1d():
    return pd.Series(np.array([12., 3., 20.]), index=pd.interval_range(start=1, periods=3))


@pytest.fixture
def some_histogram_2d():
    return pd.Series(
        np.array([5., 10., 10., 20.]),
        index=pd.MultiIndex.from_product([
            pd.interval_range(start=0, end=2),
            pd.interval_range(start=0, end=2)
        ], names=['foo', 'bar'])
    )


@pytest.fixture
def another_histogram_2d():
    return pd.Series(
        np.array([
            12., 3., 20.,
            24., 6., 40.,
            36., 9., 60.
        ]),
        index=pd.MultiIndex.from_product([
            pd.interval_range(start=0, periods=3),
            pd.interval_range(start=0, periods=3)
        ], names=['foo', 'bar'])
    )


@pytest.fixture
def some_histogram_3d():
    return pd.Series(
        np.array([
            5., 10.,
            10., 20.,
            15., 30.,
            20., 40.
        ]),
        index=pd.MultiIndex.from_product([
            pd.interval_range(start=0, end=2),
            pd.interval_range(start=2, end=4),
            pd.interval_range(start=0, periods=2)
        ], names=['foo', 'bar', 'baz'])
    )


@pytest.fixture
def hists_to_combine_1d(some_histogram_1d, another_histogram_1d):
    return [some_histogram_1d, another_histogram_1d]


def test_combine_zero_histograms(empty_histogram):
    result = hi.combine_histogram([])
    pd.testing.assert_series_equal(result, empty_histogram)


@pytest.mark.parametrize('method', ['sum', 'min', 'max', 'mean'])
def test_combine_two_empty_histograms_1d(empty_histogram, method):
    result = hi.combine_histogram([empty_histogram, empty_histogram], method=method)
    pd.testing.assert_series_equal(result, empty_histogram)


@pytest.mark.parametrize('method', ['sum', 'max'])
def test_combine_non_empty_with_empty(empty_histogram, some_histogram_1d, method):
    result = hi.combine_histogram([some_histogram_1d, empty_histogram], method=method)
    pd.testing.assert_series_equal(result, some_histogram_1d)


@pytest.mark.parametrize('method', ['sum', 'min', 'max', 'mean'])
def test_combine_empty_with_non_empty(empty_histogram, another_histogram_1d, method):
    result = hi.combine_histogram([empty_histogram, another_histogram_1d], method=method)
    pd.testing.assert_series_equal(result, another_histogram_1d)


@pytest.mark.parametrize('method, expected', [
    ('sum', [5., 22., 3., 20.]),
    ('min', [5., 10., 3., 20.]),
    ('max', [5., 12., 3., 20.]),
    ('mean', [5., 11., 3., 20.])
])
def test_combine_two_histograms_1d(some_histogram_1d, another_histogram_1d, method, expected):
    result = hi.combine_histogram([some_histogram_1d, another_histogram_1d], method=method)
    expected = pd.Series(expected, index=pd.interval_range(start=0, end=4, periods=4))
    pd.testing.assert_series_equal(result, expected)


def test_combine_1d_with_2d_histogram(some_histogram_1d, another_histogram_2d):
    with pytest.raises(ValueError, match=r"Histograms must have identical dimensions to be combined."):
        hi.combine_histogram([some_histogram_1d, another_histogram_2d])


def test_combine_1d_with_2d_histogram_duplicate_index_names(some_histogram_1d, another_histogram_2d):
    some = some_histogram_1d.copy(deep=True)
    other = another_histogram_2d.copy(deep=True)
    some.index.name = 'foo'
    other.index.names = ['foo', 'foo']
    with pytest.raises(ValueError, match=r"Histograms must have identical dimensions to be combined."):
        hi.combine_histogram([some, other])


def test_combine_2d_with_3d_histogram(some_histogram_2d, some_histogram_3d):
    with pytest.raises(ValueError, match=r"Histograms must have identical dimensions to be combined."):
        hi.combine_histogram([some_histogram_2d, some_histogram_3d])


def test_combine_2d_with_2d_histogram_inconsistent_names(some_histogram_2d):
    other = some_histogram_2d.copy(deep=True)
    other.index.names = ['foo', 'baz']
    with pytest.raises(ValueError, match=r"Histograms must have identical dimensions to be combined."):
        hi.combine_histogram([some_histogram_2d, other])


@pytest.mark.parametrize('method, expected', [
    ('sum', [17., 13., 20., 34., 26., 40., 36., 9., 60.]),
    ('min', [5., 3., 20., 10., 6., 40., 36., 9., 60.]),
    ('max', [12., 10., 20., 24., 20., 40., 36., 9., 60.])
])
def test_combine_2d_with_2d_histogram(some_histogram_2d, another_histogram_2d, method, expected):
    result = hi.combine_histogram([some_histogram_2d, another_histogram_2d], method=method)

    expected_index = pd.MultiIndex.from_product(
        [pd.interval_range(start=0, end=3, periods=3), pd.interval_range(start=0, end=3, periods=3)],
        names=['foo', 'bar']
    )
    expected = pd.Series(expected, index=expected_index)

    pd.testing.assert_series_equal(result, expected)


def test_combine_hist_sum(hists_to_combine_1d):
    expect_sum = pd.Series(np.array([5., 22., 3., 20.]), index=pd.interval_range(start=0, periods=4))
    test_sum = hi.combine_histogram(hists_to_combine_1d, method='sum')
    pd.testing.assert_series_equal(test_sum, expect_sum, check_names=False)


def test_combine_hist_min(hists_to_combine_1d):
    expect_min = pd.Series(
        np.array([5., 10., 3., 20.]),
        index=pd.interval_range(start=0, end=4, periods=4)
    )
    test_min = hi.combine_histogram(hists_to_combine_1d, method='min')
    pd.testing.assert_series_equal(test_min, expect_min, check_names=False)


def test_combine_hist_max(hists_to_combine_1d):
    expect_max = pd.Series(
        np.array([5., 12., 3., 20.]),
        index=pd.interval_range(start=0, end=4, periods=4)
    )
    test_max = hi.combine_histogram(hists_to_combine_1d, method='max')
    pd.testing.assert_series_equal(test_max, expect_max, check_names=False)


def test_combine_hist_mean(hists_to_combine_1d):
    expect_mean = pd.Series(
        np.array([5., 11., 3., 20.]),
        index=pd.interval_range(start=0, end=4, periods=4)
    )
    test_mean = hi.combine_histogram(hists_to_combine_1d, method='mean')
    pd.testing.assert_series_equal(test_mean, expect_mean, check_names=False)


def test_combine_hist_std(hists_to_combine_1d):
    expect_std = pd.Series(
        np.array([0, np.std([10, 12]), 0., 0.]),
        index=pd.interval_range(start=0, end=4, periods=4)
    )
    test_std = hi.combine_histogram(hists_to_combine_1d, method=lambda x: np.std(x))
    pd.testing.assert_series_equal(test_std, expect_std, check_names=False)


def test_combine_test_big_bin():
    hist1 = pd.Series([1, 2], index=pd.IntervalIndex.from_tuples([(0.0, 0.5), (0.5, 1.0)]))
    hist2 = pd.Series([4], index=pd.IntervalIndex.from_tuples([(0.0, 1.0)]))

    result = hi.combine_histogram([hist1, hist2], method="sum")

    expected = pd.Series(
        [1, 4, 2],
        index=pd.IntervalIndex.from_tuples([(0.0, 0.5), (0.0, 1.0), (0.5, 1.0)])
    )

    pd.testing.assert_series_equal(result, expected)




@pytest.fixture
def regular_binning():
    return pd.IntervalIndex.from_tuples([(0.0, 1.0), (1.0, 2.0), (2.0, 3.0), (3.0, 4.0)])


def test_rebin_histogram_invalid_binning(empty_histogram):
    binning = None
    with pytest.raises(TypeError, match=r"binning argument must be a pandas.IntervalIndex."):
        hi.rebin_histogram(empty_histogram, binning)


def test_rebin_invalid_without_interval_index():
    binning = pd.IntervalIndex.from_tuples([(0.0, 1.0), (1.0, 2.0), (0.0, 1.0), (1.0, 2.0)])
    hist = pd.Series(dtype=np.float64)
    with pytest.raises(TypeError, match=r"histogram needs to have an IntervalIndex."):
        hi.rebin_histogram(hist, binning)


def test_rebin_histogram_non_monotonic_binning(empty_histogram):
    binning = pd.IntervalIndex.from_tuples([(0.0, 1.0), (1.0, 2.0), (0.0, 1.0), (1.0, 2.0)])
    with pytest.raises(ValueError, match=r"binning index must be monotonic increasing without overlaps."):
        hi.rebin_histogram(empty_histogram, binning)


def test_rebin_histogram_binning_decreasing(empty_histogram, regular_binning):
    with pytest.raises(ValueError, match=r"binning index must be monotonic increasing without overlaps."):
        hi.rebin_histogram(empty_histogram, regular_binning[::-1])


def test_rebin_histogram_binning_with_overlaps(empty_histogram):
    binning = pd.IntervalIndex.from_tuples([(0.0, 1.0), (1.0, 2.0), (2.5, 3.0), (2.5, 3.5)])
    with pytest.raises(ValueError, match=r"binning index must be monotonic increasing without overlaps."):
        hi.rebin_histogram(empty_histogram, binning)


def test_rebin_histogram_gapped_binning(empty_histogram):
    binning = pd.IntervalIndex.from_tuples([(0.0, 1.0), (1.0, 2.0), (2.0, 3.0), (4.0, 5.0)])
    with pytest.raises(ValueError, match=r"binning index must not have gaps."):
        hi.rebin_histogram(empty_histogram, binning)


@pytest.mark.parametrize('hist', [
    (pd.Series([1.0, 1.0], index=pd.IntervalIndex.from_tuples([(0.3, 0.7), (5.6, 6.7)]))),
    (pd.Series([1.0, 1.0], index=pd.IntervalIndex.from_tuples([(0.3, 0.7), (-6.6, -5.7)]))),
])
def test_rebin_histogram_insufficient_binning(hist):
    binning = pd.IntervalIndex.from_tuples([(0.0, 1.0), (1.0, 2.0)])
    with pytest.warns(RuntimeWarning, match=r"histogram is partly out of binning. "
                      "This information will be lost!"):
        hi.rebin_histogram(hist, binning)


def test_rebin_empty_histogram(empty_histogram, regular_binning):
    rebinned = hi.rebin_histogram(empty_histogram, regular_binning)

    pd.testing.assert_series_equal(rebinned, pd.Series(0.0, index=regular_binning))
    pd.testing.assert_index_equal(rebinned.index, regular_binning)


def test_rebin_histogram_name_empty_histogram(empty_histogram, regular_binning):
    empty_histogram.name = "empty_histogram"
    rebinned = hi.rebin_histogram(empty_histogram, regular_binning)

    assert rebinned.name == "empty_histogram"


def test_rebin_histogram_name_non_empty_histogram(regular_binning):
    histogram = pd.Series([1.0], index=pd.IntervalIndex.from_tuples([(0.2, 0.4)]), name="foo")
    rebinned = hi.rebin_histogram(histogram, regular_binning)

    assert rebinned.name == "foo"


def test_rebin_histogram_index_name_empty_histogram(empty_histogram, regular_binning):
    empty_histogram.index.name = "index_name"
    rebinned = hi.rebin_histogram(empty_histogram, regular_binning)

    assert rebinned.index.name == "index_name"


def test_rebin_histogram_index_name_non_empty_histogram(regular_binning):
    histogram = pd.Series(
        [1.0], index=pd.IntervalIndex.from_tuples([(0.2, 0.4)], name="index_name")
    )
    rebinned = hi.rebin_histogram(histogram, regular_binning)

    assert rebinned.index.name == "index_name"


@pytest.mark.parametrize('histogram, expected',[
    (
        pd.Series([1.0], index=pd.IntervalIndex.from_tuples([(0.2, 0.4)])),
        [1.0, 0.0, 0.0, 0.0]
    ),
    (
        pd.Series([2.0], index=pd.IntervalIndex.from_tuples([(0.2, 0.4)])),
        [2.0, 0.0, 0.0, 0.0]
    ),
    (
        pd.Series([2.0], index=pd.IntervalIndex.from_tuples([(1.2, 1.4)])),
        [0.0, 2.0, 0.0, 0.0]
    ),
    (
        pd.Series([2.0, 1.0], index=pd.IntervalIndex.from_tuples([(1.2, 1.4), (0.2, 0.4)])),
        [1.0, 2.0, 0.0, 0.0]
    ),
    (
        pd.Series([2.0, 1.0], index=pd.IntervalIndex.from_tuples([(0.5, 0.9), (0.2, 0.4)])),
        [3.0, 0.0, 0.0, 0.0]
    ),
    (
        pd.Series([2.0], index=pd.IntervalIndex.from_tuples([(0.5, 1.5)])),
        [1.0, 1.0, 0.0, 0.0]
    ),
    (
        pd.Series([2.0, 3.0], index=pd.IntervalIndex.from_tuples([(0.5, 1.5), (3.4, 3.6)])),
        [1.0, 1.0, 0.0, 3.0]
    ),
    (
        pd.Series([2.0, 4.0], index=pd.IntervalIndex.from_tuples([(0.5, 1.5), (1.5, 3.5)])),
        [1.0, 2.0, 2.0, 1.0]
    ),
])
def test_rebin_histogram(histogram, expected, regular_binning):
    rebinned = hi.rebin_histogram(histogram, regular_binning)

    pd.testing.assert_series_equal(rebinned, pd.Series(expected, index=regular_binning))


def test_rebin_histogram_nan_default(regular_binning):
    histogram = pd.Series([1.0], index=pd.IntervalIndex.from_tuples([(0.2, 0.4)]))
    expected = pd.Series([1.0, np.nan, np.nan, np.nan], index=regular_binning)

    rebinned = hi.rebin_histogram(histogram, regular_binning, nan_default=True)

    pd.testing.assert_series_equal(rebinned, expected)


@pytest.mark.parametrize('original_binning, binnum, expected', [
    ([(0.0, 1.0)], 4, [(0.0, 0.25), (0.25, 0.5), (0.5, 0.75), (0.75, 1.0)]),
    ([(0.0, 1.0)], 2, [(0.0, 0.5), (0.5, 1.0)]),
    ([(1.0, 2.0)], 4, [(1.0, 1.25), (1.25, 1.5), (1.5, 1.75), (1.75, 2.0)]),
    ([(1.0, 2.0)], 2, [(1.0, 1.5), (1.5, 2.0)]),
])
def test_rebin_histogram_n_bins(original_binning, binnum, expected):
    histogram = pd.Series([1.0], pd.IntervalIndex.from_tuples(original_binning))
    rebinned = hi.rebin_histogram(histogram, binnum)

    expected = pd.IntervalIndex.from_tuples(expected)
    pd.testing.assert_index_equal(rebinned.index, expected)


def test_rebin_histogram_additional_index():
    hi_index = pd.IntervalIndex.from_tuples([(0.0, 0.5), (0.5, 1.0)], name='hi')
    other_index = pd.Index([1, 2], name='foo')
    index = pd.MultiIndex.from_product([hi_index, other_index])

    histogram = pd.Series([1.0, 2.0, 3.0, 4.0], index=index)

    expected_hi_index = pd.Index(
        pd.interval_range(start=0, end=1, periods=4),
        name='hi'
    )
    expected_index = pd.MultiIndex.from_product([expected_hi_index, other_index])
    expected_histogram = pd.Series([0.5, 1.0, 0.5, 1.0, 1.5, 2.0, 1.5, 2.0], index=expected_index)

    result = hi.rebin_histogram(histogram, 4)

    pd.testing.assert_series_equal(result.sort_index(), expected_histogram)


def test_rebin_histogram_additional_index_given_rebin():
    hi_index = pd.IntervalIndex.from_tuples([(0.0, 0.5), (0.5, 1.0)], name='hi')
    other_index = pd.Index([1, 2], name='foo')
    index = pd.MultiIndex.from_product([hi_index, other_index])

    histogram = pd.Series([1.0, 2.0, 3.0, 4.0], index=index)

    new_hi_index = pd.Index(
        pd.interval_range(start=0, end=1, periods=4),
        name='hi'
    )
    expected_index = pd.MultiIndex.from_product([new_hi_index, other_index])
    expected_histogram = pd.Series([0.5, 1.0, 0.5, 1.0, 1.5, 2.0, 1.5, 2.0], index=expected_index)

    result = hi.rebin_histogram(histogram, new_hi_index)

    pd.testing.assert_series_equal(result.sort_index(), expected_histogram)


def test_rebin_histogram_2d():
    idx_x = pd.interval_range(0, 1, 2)
    idx_y = pd.interval_range(0, 10, 2)
    histogram = pd.Series(1.0, index=pd.MultiIndex.from_product([idx_x, idx_y], names=['foo', 'bar']))

    idx_x = pd.interval_range(0, 1, 4)
    idx_y = pd.interval_range(0, 10, 4)

    target_idx = pd.MultiIndex.from_product([idx_x, idx_y], names=['foo', 'bar'])

    expected = pd.Series(0.25, index=target_idx)

    result = hi.rebin_histogram(histogram, 4)

    assert result.sum() == histogram.sum()
    pd.testing.assert_series_equal(result, expected)


def test_rebin_histogram_2d_not_rebin_fill():
    idx_x = pd.interval_range(0., 1., 2)
    idx_y = pd.interval_range(0., 10., 2)
    histogram = pd.Series([1., 2., 3., 4.], index=pd.MultiIndex.from_product([idx_x, idx_y], names=['foo', 'bar']))

    idx_x = pd.interval_range(0., 2., 4)
    idx_y = pd.interval_range(0., 20., 4)

    target_idx = pd.MultiIndex.from_product([idx_x, idx_y], names=['foo', 'bar'])

    expected = pd.Series([1.0, 2.0, np.nan, np.nan, 3.0, 4.0] + [np.nan]*10, index=target_idx)

    result = hi.rebin_histogram(histogram, target_idx, nan_default=True)
    pd.testing.assert_series_equal(result, expected)


def test_rebin_histogram_2d_given_rebin():
    idx_x = pd.interval_range(0, 1, 2)
    idx_y = pd.interval_range(0, 10, 2)
    histogram = pd.Series(1.0, index=pd.MultiIndex.from_product([idx_x, idx_y], names=['foo', 'bar']))

    idx_x = pd.interval_range(0, 1, 4)
    idx_y = pd.interval_range(0, 10, 4)

    target_idx = pd.MultiIndex.from_product([idx_x, idx_y], names=['foo', 'bar'])

    expected = pd.Series(0.25, index=target_idx)

    result = hi.rebin_histogram(histogram, target_idx)

    assert result.sum() == histogram.sum()
    pd.testing.assert_series_equal(result, expected)


def test_rebin_histogram_3d():
    idx_x = pd.interval_range(0., 1., 2)
    idx_y = pd.interval_range(0., 10., 2)
    idx_z = pd.interval_range(0., 100., 2)
    histogram = pd.Series(2.0, index=pd.MultiIndex.from_product(
        [idx_x, idx_y, idx_z],
        names=['foo', 'bar', 'baz']
    ))

    idx_x = pd.interval_range(0., 1., 4)
    idx_y = pd.interval_range(0., 10., 4)
    idx_z = pd.interval_range(0., 100., 4)

    target_idx = pd.MultiIndex.from_product(
        [idx_x, idx_y, idx_z],
        names=['foo', 'bar', 'baz']
    )

    expected = pd.Series(0.25, index=target_idx)

    result = hi.rebin_histogram(histogram, 4)

    assert result.sum() == histogram.sum()
    pd.testing.assert_series_equal(result, expected)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__


import numpy as np
import pandas as pd

from .general import AbstractRecorder


class LoopValueRecorder(AbstractRecorder):
    """Rainflow recorder that collects the loop values."""

    def __init__(self):
        """Instantiate a LoopRecorder."""
        super().__init__()
        self._values_from = np.zeros((0,))
        self._values_to = np.zeros((0,))

    @property
    def values_from(self):
        """1-D float array containing the values from which the loops start."""
        return self._values_from

    @property
    def values_to(self):
        """1-D float array containing the values the loops go to before turning back."""
        return self._values_to

    @property
    def collective(self):
        """The overall collective recorded as :class:`pandas.DataFrame`.

        The columns are named ``from``, ``to``.
        """
        return pd.DataFrame({'from': self._values_from, 'to': self._values_to})

    def record_values(self, values_from, values_to):
        """Record the loop values."""
        self._values_from = np.append(self._values_from, values_from)
        self._values_to = np.append(self._values_to, values_to)

    def histogram_numpy(self, bins=10):
        """Calculate a histogram of the recorded values into a plain numpy.histogram2d.

        Parameters
        ----------
        bins : int or array_like or [int, int] or [array, array], optional
            The bin specification (see numpy.histogram2d)

        Returns
        -------
        H : ndarray, shape(nx, ny)
            The bi-dimensional histogram of samples (see numpy.histogram2d)
        xedges : ndarray, shape(nx+1,)
            The bin edges along the first dimension.
        yedges : ndarray, shape(ny+1,)
            The bin edges along the second dimension.
        """
        return np.histogram2d(self._values_from, self._values_to, bins)

    def histogram(self, bins=10):
        """Calculate a histogram of the recorded values into a :class:`pandas.Series`.

        An interval index is used to index the bins.

        Parameters
        ----------
        bins : int or array_like or [int, int] or [array, array], optional
            The bin specification (see numpy.histogram2d)

        Returns
        -------
        pandas.Series
            A pandas.Series using a multi interval index in order to
            index data point for a given from/to value pair.
        """
        hist, fr, to = self.histogram_numpy(bins)
        index_fr = pd.IntervalIndex.from_breaks(fr)
        index_to = pd.IntervalIndex.from_breaks(to)

        mult_idx = pd.MultiIndex.from_product([index_fr, index_to], names=['from', 'to'])
        return pd.Series(data=hist.flatten(), index=mult_idx)


class FullRecorder(LoopValueRecorder):
    """Rainflow recorder that collects the loop values and the loop index.

    Same functionality like :class:`.LoopValueRecorder` but additionally
    collects the loop index.
    """

    def __init__(self):
        """Instantiate a FullRecorder."""
        super().__init__()
        self._index_from = np.array([], dtype=np.uintp)
        self._index_to = np.array([], dtype=np.uintp)

    @property
    def index_from(self):
        """1-D int array containing the index to the samples from which the loops start."""
        return self._index_from

    @property
    def index_to(self):
        """1-D int array containing the index to the samples the loops go to before turning back."""
        return self._index_to

    @property
    def collective(self):
        """The overall collective recorded as :class:`pandas.DataFrame`.

        The columns are named ``from``, ``to``, ``index_from``, ``index_to``.
        """
        return pd.DataFrame({
            'from': self._values_from,
            'to': self._values_to,
            'index_from': self._index_from,
            'index_to': self._index_to
        })

    def record_index(self, index_from, index_to):
        """Record the index."""
        self._index_from = np.concatenate(
            (self._index_from, np.asarray(index_from, dtype=np.uintp))
        )
        self._index_to = np.concatenate(
            (self._index_to, np.asarray(index_to, dtype=np.uintp))
        )



# Copyright (c) 2019-2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import scipy.stats as stats

from .likelihood import Likelihood
from .pearl_chain import PearlChainProbability
import pylife.utils.functions as functions
from . import FatigueData, determine_fractures
import warnings

class Elementary:
    """Base class to analyze SN-data.

    The common base class for all SN-data analyzers calculates the first
    estimation of a Wöhler curve in the finite zone of the SN-data. It
    calculates the slope `k`, the fatigue limit `SD`, the transition cycle
    number `ND` and the scatter in load direction `1/TN`.

    The result is just meant to be a first guess. Derived classes are supposed
    to use those first guesses as starting points for their specific
    analysis. For that they should implement the method `_specific_analysis()`.
    """

    def __init__(self, fatigue_data):
        """The constructor.

        Parameters
        ----------
        fatigue_data : pd.DataFrame or FatigueData
           The SN-data to be analyzed.
        """
        self._fd = self._get_fatigue_data(fatigue_data)
        self._lh = self._get_likelihood()

    def _get_likelihood(self):
        return Likelihood(self._fd)

    def _get_fatigue_data(self, fatigue_data):
        if isinstance(fatigue_data, pd.DataFrame):
            if hasattr(fatigue_data, "fatigue_data"):
                params = fatigue_data.fatigue_data
            else:
                params = determine_fractures(fatigue_data).fatigue_data
        elif isinstance(fatigue_data, FatigueData):
            params = fatigue_data
        else:
            raise ValueError("fatigue_data of type {} not understood: {}".format(type(fatigue_data), fatigue_data))
        params = params.irrelevant_runouts_dropped()

        return params

    def analyze(self, **kwargs):
        """Analyze the SN-data.

        Parameters
        ----------
        \*\*kwargs : kwargs arguments
            Arguments to be passed to the derived class
        """
        if len(self._fd.load.unique()) < 2:
            raise ValueError("Need at least two different load levels in the finite zone to do a Wöhler slope analysis.")
        elif len(self._fd.finite_zone.load.unique()) < 2:
            warnings.warn(UserWarning("Need at least two different load levels in the finite zone to do a Wöhler slope analysis."
                                      " "))
            if len(self._fd.finite_zone.load.unique()) == 1:
                wc = pd.Series({
                    'k_1': np.nan,
                    'ND': np.nan,
                    'SD': np.nan,
                    'TN': np.nan,
                    'TS': np.nan
                    })
            else:
                wc = pd.Series({
                    'k_1': np.inf,
                    'ND': np.nan,
                    'SD': np.nan,
                    'TN': 1.0,
                    'TS': np.nan
                    })
            wc = self._specific_analysis(wc, **kwargs)
            wc['failure_probability'] = 0.5
            return wc

        self._finite_fractures = self._fd.finite_zone.loc[self._fd.finite_zone.fracture == True]
        wc = self._common_analysis()
        wc = self._specific_analysis(wc, **kwargs)
        self.__calc_bic(wc)
        wc['failure_probability'] = 0.5

        return wc

    def _common_analysis(self):
        self._slope, self._lg_intercept = self._fit_slope()
        TN, TS = self._pearl_chain_method()
        return pd.Series({
            'k_1': -self._slope,
            'ND': self._transition_cycles(self._fd.finite_infinite_transition),
            'SD': self._fd.finite_infinite_transition,
            'TN': TN,
            'TS': TS
        })

    def _specific_analysis(self, wc):
        return wc

    def bayesian_information_criterion(self):
        """The Bayesian Information Criterion

        Bayesian Information Criterion is a criterion for model selection among
        a finite set of models; the model with the lowest BIC is preferred.
        https://www.statisticshowto.datasciencecentral.com/bayesian-information-criterion/

        Basically the lower the better the fit.
        """
        if not hasattr(self,"_bic"):
            raise ValueError("BIC value undefined. Analysis has not been conducted.")
        return self._bic

    def pearl_chain_estimator(self):
        return self._pearl_chain_estimator

    def __calc_bic(self, wc):
        '''         '''
        param_num = 5  # SD, TS, k_1, ND, TN
        log_likelihood = self._lh.likelihood_total(wc['SD'], wc['TS'], wc['k_1'], wc['ND'], wc['TN'])
        self._bic = (-2 * log_likelihood) + (param_num * np.log(self._fd.num_tests))

    def _fit_slope(self):
        slope, lg_intercept, _, _, _ = stats.linregress(np.log10(self._finite_fractures.load),
                                                        np.log10(self._finite_fractures.cycles))

        return slope, lg_intercept

    def _transition_cycles(self, finite_infinite_transition):
        # FIXME Elementary means finite_infinite_transition == 0 -> np.inf
        if finite_infinite_transition == 0:
            finite_infinite_transition = 0.1
        return 10**(self._lg_intercept + self._slope * (np.log10(finite_infinite_transition)))

    def _pearl_chain_method(self):
        self._pearl_chain_estimator = PearlChainProbability(self._finite_fractures, self._slope)

        TN = functions.std_to_scattering_range(1./self._pearl_chain_estimator.slope)
        TS = TN**(1./-self._slope)

        return TN, TS



# Copyright (c) 2019-2024 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Vishnu Pradeep"
__maintainer__ = "Johannes Mueller"


import numpy as np
from pylife.rainflow_ext import fourpoint_loop

from .general import AbstractDetector


class FourPointDetector(AbstractDetector):
    """Implements four point rainflow counting algorithm.

    .. jupyter-execute::

        from pylife.stress.timesignal import TimeSignalGenerator
        import pylife.stress.rainflow as RF

        ts = TimeSignalGenerator(10, {
            'number': 50,
            'amplitude_median': 1.0, 'amplitude_std_dev': 0.5,
            'frequency_median': 4, 'frequency_std_dev': 3,
            'offset_median': 0, 'offset_std_dev': 0.4}, None, None).query(10000)

        rfc = RF.FourPointDetector(recorder=RF.LoopValueRecorder())
        rfc.process(ts)

        rfc.recorder.collective

    Alternatively you can ask the recorder for a histogram matrix:

    .. jupyter-execute::

        rfc.recorder.histogram(bins=16)

    We take four turning points into account to detect closed hysteresis loops.

    Consider four consecutive peak/valley points say, A, B, C, and D  If B and C are
    contained within A and B, then a cycle is counted from B to C; otherwise no cycle is
    counted.

    i.e, If ``X ≥ Y AND Z ≥ Y`` then a cycle exist ``FROM = B`` and ``TO = C``
    where, ranges ``X = |D–C|``, ``Y = |C–B|``, and ``Z = |B–A|``

    ::

        Load -----------------------------
        |        x B               F x
        --------/-\-----------------/-----
        |      /   \   x D         /
        ------/-----\-/-\---------/-------
        |    /     C x   \       /
        --\-/-------------\-----/---------
        |  x A             \   /
        --------------------\-/-----------
        |                    x E
        ----------------------------------
        |              Time

    So, if a cycle exsist from B to C then delete these peaks from the turns array
    and perform next iteration by joining A&D else if no cylce exsists, then B would
    be the next strarting point.
    """

    def __init__(self, recorder):
        """Instantiate a FourPointDetector.

        Parameters
        ----------
        recorder : subclass of :class:`.AbstractRecorder`
            The recorder that the detector will report to.
        """
        super().__init__(recorder)

    def process(self, samples):
        """Process a sample chunk.

        Parameters
        ----------
        samples : array_like, shape (N, )
            The samples to be processed

        Returns
        -------
        self : FourPointDetector
            The ``self`` object so that processing can be chained
        """

        samples = np.asarray(samples)
        residuals = samples[:1] if self._residuals.size == 0 else self._residuals[:-1]

        turns_index, turns_values = self._new_turns(samples)

        turns_np = np.concatenate((residuals, turns_values, samples[-1:]))
        turns_index = np.concatenate((self._residual_index, turns_index.astype(np.uintp)))

        (
            from_vals,
            to_vals,
            from_index,
            to_index,
            residual_index
        ) = fourpoint_loop(turns_np, turns_index)

        self._recorder.record_values(from_vals, to_vals)
        self._recorder.record_index(from_index, to_index)

        self._residuals = turns_np[residual_index]
        self._residual_index = turns_index[residual_index[:-1]]
        self._recorder.report_chunk(len(samples))

        return self



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

from abc import ABCMeta, abstractmethod
import warnings
import numpy as np
import pandas as pd


def find_turns(samples):
    """Find the turning points in a sample chunk.

    Parameters
    ----------
    samples : 1D numpy.ndarray
        the sample chunk

    Returns
    -------
    index : 1D numpy.ndarray
        the indeces where sample has a turning point
    turns : 1D numpy.ndarray
        the values of the turning points

    Notes
    -----
    In case of plateaus i.e. multiple directly neighbored samples with exactly
    the same values, building a turning point together, the first sample of the
    plateau is indexed.


    Warnings
    --------
    Any ``NaN`` values are dropped from the input signal before processing it
    and will thus also not appear in the turns.  In those cases a warning is
    issued.  The reason for this is, that if the ``NaN`` appears next to an
    actual turning point the turning point is no longer detected which will
    lead to an underestimation of the damage sum later in the damage
    calculation.  Generally you should not have ``NaN`` values in your signal.
    If you do, it would be a good idea to clean them out before the rainflow
    detection.

    """

    def clean_nans(samples):
        nans = pd.isna(samples)
        if nans.any():
            warnings.warn(UserWarning("At least one NaN like value has been dropped from the input signal."))
            return samples[~nans], nans
        return samples, None

    def correct_turns_by_nans(index, nans):
        if nans is None:
            return
        nan_positions = np.where(nans)[0]
        for nan_pos in nan_positions:
            index[index >= nan_pos] += 1

    def plateau_turns(diffs):
        plateau_turns = np.zeros_like(diffs, dtype=np.bool_)[1:]
        duplicates = np.array(diffs == 0, dtype=np.int8)

        if duplicates.any():
            edges = np.diff(duplicates)
            dups_starts = np.where(edges > 0)[0]
            dups_ends = np.where(edges < 0)[0]
            if len(dups_starts) and len(dups_ends):
                cut_ends = dups_ends[0] < dups_starts[0]
                cut_starts = dups_starts[-1] > dups_ends[-1]
                if cut_ends:
                    dups_ends = dups_ends[1:]
                if cut_starts:
                    dups_starts = dups_starts[:-1]
                plateau_turns[dups_starts[np.where(diffs[dups_starts] * diffs[dups_ends+1] < 0)]] = True

        return plateau_turns

    samples, nans = clean_nans(samples)

    diffs = np.diff(samples)
    peak_turns = diffs[:-1] * diffs[1:] < 0.0

    index = np.where(np.logical_or(peak_turns, plateau_turns(diffs)))[0] + 1

    turns_values = samples[index]
    correct_turns_by_nans(index, nans)

    return index, turns_values


class AbstractDetector(metaclass=ABCMeta):
    """The common base class for rainflow detectors.

    Subclasses implementing a specific rainflow counting algorithm are supposed
    to implement a method ``process()`` that takes the signal samples as a
    parameter, and reports all the hysteresis loop limits to ``self._recorder``
    using its ``record_values()`` method of.

    Some detectors also report the index of the loop limiting samples to the
    recorder using its ``record_index()`` method. Those detectors should also
    report the size of each processed sample chunk to the recorder using
    ``report_chunk()``.

    The ``process()`` method is supposed return ``self`` and to be implemented
    in a way, that the result is independent of the sample chunksize, so
    ``dtor.process(signal)`` should be equivalent to
    ``dtor.process(signal[:n]).process(signal[n:])`` for any 0 < n < signal
    length.

    Should usually only be instantiated by a sublacsse's ``__init__()`` using
    ``super().__init__()``.
    """

    def __init__(self, recorder):
        """Instantiate an AbstractDetector.

        Parameters
        ----------
        recorder : subclass of :class:`.AbstractRecorder`
            The recorder that the detector will report to.
        """
        self._sample_tail = np.array([])
        self._recorder = recorder
        self._head_index = 0
        self._residual_index = np.array([0], dtype=np.uintp)
        self._residuals = np.array([])

    @property
    def residuals(self):
        """The residual turning points of the time signal so far.

        The residuals are the loops not (yet) closed.
        """
        return self._residuals

    @property
    def residual_index(self):
        """The index of the residual turning points of the time signal so far."""
        return np.append(self._residual_index, self._head_index - 1)

    @property
    def recorder(self):
        """The recorder instance the detector is reporting to."""
        return self._recorder

    @abstractmethod
    def process(self, samples):
        """Process a sample chunk.

        Parameters
        ----------
        samples : array_like, shape (N, )
            The samples to be processed

        Returns
        -------
        self : instance of the subclass
            The ``self`` object so that processing can be chained

        Notes
        -----
        Must be implemented by subclasses.
        """
        return self

    def _new_turns(self, samples):
        """Provide new turning points for the next chunk.

        Parameters
        ----------
        samples : 1-D array of float
            The samples of the chunk to be processed

        Returns
        -------
        turn_index : 1-D array of int
            The global index of the turning points of the chunk to be processed
        turn_values : 1-D array of float
            The values of the turning points

        Notes
        -----
        This method can be called by the ``process()`` implementation of
        subclasses. The sample tail i.e. the samples after the last turning
        point of the chunk are stored and prepended to the samples of the next
        call.
        """
        sample_len = len(samples)
        samples = np.concatenate((self._sample_tail, samples))
        turn_index, turn_values = find_turns(samples)
        if turn_index.size > 0:
            old_sample_tail_length = len(self._sample_tail)
            self._sample_tail = samples[turn_index[-1]:]
            turn_index += self._head_index - old_sample_tail_length
        else:
            self._sample_tail = samples

        self._head_index += sample_len

        return turn_index, turn_values


class AbstractRecorder:
    """A common base class for rainflow recorders.

    Subclasses implementing a rainflow recorder are supposed to implement the
    following methods:

    * ``record_values()``
    * ``record_index()``
    """

    def __init__(self):
        """Instantiate an AbstractRecorder."""
        self._chunks = np.array([], dtype=np.int64)

    @property
    def chunks(self):
        """The limits index of the chunks processed so far.

        Note
        ----
        The first chunk limit is the length of the first chunk, so identical to
        the index to the first sample of the second chunk, if a second chunk
        exists.
        """
        return self._chunks

    def report_chunk(self, chunk_size):
        """Report a chunk.

        Parameters
        ----------
        chunk_size : int
            The length of the chunk previously processed by the detector.

        Note
        ----
        Should be called by the detector after the end of ``process()``.
        """
        self._chunks = np.append(self._chunks, chunk_size)

    def chunk_local_index(self, global_index):
        """Transform the global index to an index valid in a certain chunk.

        Parameters
        ----------
        global_index : array-like int
            The global index to be transformed.

        Returns
        -------
        chunk_number : array of ints
            The number of the chunk the indexed sample is in.
        chunk_local_index : array of ints
            The index of the sample in its chunk.
        """
        chunk_index = np.insert(np.cumsum(self._chunks), 0, 0)
        chunk_num = np.searchsorted(chunk_index, global_index, side='right') - 1

        return chunk_num, global_index - chunk_index[chunk_num]

    def record_values(self, values_from, values_to):  # pragma: no cover
        """Report hysteresis loop values to the recorder.

        Parameters
        ----------
        values_from : list of floats
            The sample values where the hysteresis loop starts from.

        values_to : list of floats
            The sample values where the hysteresis loop goes to and turns back from.

        Note
        ----
        Default implementation does nothing. Can be implemented by recorders
        interested in the hysteresis loop values.
        """
        pass

    def record_index(self, indeces_from, indeces_to):  # pragma: no cover
        """Record hysteresis loop index to the recorder.

        Parameters
        ----------
        indeces_from : list of ints
            The sample indeces where the hysteresis loop starts from.

        indeces_to : list of ints
            The sample indeces where the hysteresis loop goes to and turns back from.

        Note
        ----
        Default implementation does nothing. Can be implemented by recorders
        interested in the hysteresis loop values.
        """
        pass



import pytest
import numpy as np
import re

from pylife.materiallaws import HookesLaw1d
from pylife.materiallaws import HookesLaw2dPlaneStrain
from pylife.materiallaws import HookesLaw2dPlaneStress
from pylife.materiallaws import HookesLaw3d


# HookesLaw1d
@pytest.fixture
def hookeslaw1d():
    return HookesLaw1d(E=2.)


def test_hookeslaw1d_init(hookeslaw1d):
    assert hookeslaw1d._E == 2.


def test_hookeslaw1d_properties(hookeslaw1d):
    assert hookeslaw1d.E == hookeslaw1d._E


data1D = np.array([
    [0.0, 0.0],
    [1.0, 1./2],
    [2.0, 1.],
    [3.0, 3./2.],
    [-1.0, -1./2],
    [-2.0, -1.],
    [-3.0, -3./2.]
])


@pytest.mark.parametrize('stress, expected', map(tuple, data1D))
def test_hookeslaw1d_strain_scalar(hookeslaw1d, stress, expected):
    np.testing.assert_approx_equal(hookeslaw1d.strain(stress), expected, significant=5)


@pytest.mark.parametrize('expected, strain', map(tuple, data1D))
def test_hookeslaw1d_stress_scalar(hookeslaw1d, expected, strain):
    np.testing.assert_approx_equal(hookeslaw1d.stress(strain), expected, significant=5)


@pytest.mark.parametrize('stress, expected', [(data1D[:, 0], data1D[:, 1])])
def test_hookeslaw1d_strain_array(hookeslaw1d, stress, expected):
    np.testing.assert_allclose(hookeslaw1d.strain(stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(data1D[:, 0], data1D[:, 1])])
def test_hookeslaw1d_stress_array(hookeslaw1d, expected, strain):
    np.testing.assert_allclose(hookeslaw1d.stress(strain), expected, rtol=1e-5)


@pytest.fixture
def hookeslaw1d_real():
    return HookesLaw1d(E=210.5e9)


def test_hookeslaw1d_real_init(hookeslaw1d_real):
    assert hookeslaw1d_real._E == 210.5e9


def test_hookeslaw1d_real_properties(hookeslaw1d_real):
    assert hookeslaw1d_real.E == hookeslaw1d_real._E


data1D_real = np.array([
    [0.0, 0.0],
    [100e6, 475.059e-6],
    [500e6, 2375.296e-6],
    [900e6, 4275.553e-6],
    [-100e6, -475.059e-6],
    [-500e6, -2375.296e-6],
    [-900e6, -4275.553e-6]
])


@pytest.mark.parametrize('stress, expected', map(tuple, data1D_real))
def test_hookeslaw1d_real_strain_scalar(hookeslaw1d_real, stress, expected):
    np.testing.assert_approx_equal(hookeslaw1d_real.strain(stress), expected, significant=5)


@pytest.mark.parametrize('expected, strain', map(tuple, data1D_real))
def test_hookeslaw1d_real_stress_scalar(hookeslaw1d_real, expected, strain):
    np.testing.assert_approx_equal(hookeslaw1d_real.stress(strain), expected, significant=5)


@pytest.mark.parametrize('stress, expected', [(data1D_real[:, 0], data1D_real[:, 1])])
def test_hookeslaw1d_real_strain_array(hookeslaw1d_real, stress, expected):
    np.testing.assert_allclose(hookeslaw1d_real.strain(stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(data1D_real[:, 0], data1D_real[:, 1])])
def test_hookeslaw1d_real_stress_array(hookeslaw1d_real, expected, strain):
    np.testing.assert_allclose(hookeslaw1d_real.stress(strain), expected, rtol=1e-5)


# HookesLaw2dPlaneStress
@pytest.fixture
def hookeslaw2dplainstress():
    return HookesLaw2dPlaneStress(E=2., nu=0.1)


@pytest.mark.parametrize('nu', [-1.1, 0.6])
def test_hookeslaw2dplainstress_init(nu):
    with pytest.raises(ValueError,
                       match=re.escape('Possion\'s ratio nu is %.2f but must be -1 <= nu <= 1./2.' % nu)):
        HookesLaw2dPlaneStress(E=2., nu=nu)


def test_hookeslaw2dplainstress_wrongshapestress(hookeslaw2dplainstress):
    with pytest.raises(ValueError,
                       match=re.escape('Components\' shape is not consistent.')):
        hookeslaw2dplainstress.stress(e11=np.zeros((1, 2)), e22=np.zeros((1, 2)), g12=np.zeros((2, 1)))


def test_hookeslaw2dplainstress_wrongshapestrain(hookeslaw2dplainstress):
    with pytest.raises(ValueError,
                       match=re.escape('Components\' shape is not consistent.')):
        hookeslaw2dplainstress.strain(s11=np.zeros((1, 2)), s22=np.zeros((1, 2)), s12=np.zeros((2, 1)))


def test_hookeslaw2dplainstress_init(hookeslaw2dplainstress):
    assert hookeslaw2dplainstress._E == 2.
    assert hookeslaw2dplainstress._nu == 0.1


def test_hookeslaw2dplainstress_properties(hookeslaw2dplainstress):
    assert hookeslaw2dplainstress.E == hookeslaw2dplainstress._E
    assert hookeslaw2dplainstress.nu == hookeslaw2dplainstress._nu
    np.testing.assert_approx_equal(hookeslaw2dplainstress.G, 0.9090909, significant=5)
    np.testing.assert_approx_equal(hookeslaw2dplainstress.K, 0.8333333, significant=5)


# (s11, s22, s33, s12), (e11, e22, e33, g12)
data2Dstress = np.array([
    [[0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]],
    [[1.0, 0.0, 0.0, 0.0], [0.5, -0.05, -0.05, 0.0]],
    [[0.0, 1.0, 0.0, 0.0], [-0.05, 0.5, -0.05, 0.0]],
    [[0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 1.1]],
    [[1.0, 2.0, 0.0, 0.5], [0.4, 0.95, -0.15, 0.55]],
    [[-0.5, 1.0, 0.0, 2.0], [-0.3, 0.525, -0.025, 2.2]],
    [[0.5, -0.5, 0.0, 1.0], [0.275, -0.275, 0.0, 1.1]]
])


@pytest.mark.parametrize('stress, expected', map(tuple, data2Dstress))
def test_hookeslaw2dplainstress_strain_scalar(hookeslaw2dplainstress, stress, expected):
    stress = dict(s11=stress[0], s22=stress[1], s12=stress[3])
    np.testing.assert_allclose(hookeslaw2dplainstress.strain(**stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', map(tuple, data2Dstress))
def test_hookeslaw2dplainstress_stress_scalar(hookeslaw2dplainstress, expected, strain):
    strain = dict(e11=strain[0], e22=strain[1], g12=strain[3])
    np.testing.assert_allclose(hookeslaw2dplainstress.stress(**strain), expected[[0, 1, 3]], rtol=1e-5)


@pytest.mark.parametrize('stress, expected', [(data2Dstress[:, 0], data2Dstress[:, 1])])
def test_hookeslaw2dplainstress_strain_array(hookeslaw2dplainstress, stress, expected):
    stress = dict(s11=stress[:, 0], s22=stress[:, 1], s12=stress[:, 3])
    np.testing.assert_allclose(np.array(hookeslaw2dplainstress.strain(**stress)).T, expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(data2Dstress[:, 0], data2Dstress[:, 1])])
def test_hookeslaw2dplainstress_stress_array(hookeslaw2dplainstress, expected, strain):
    strain = dict(e11=strain[:, 0], e22=strain[:, 1], g12=strain[:, 3])
    np.testing.assert_allclose(np.array(hookeslaw2dplainstress.stress(**strain)).T, expected[:, (0, 1, 3)], rtol=1e-5)


@pytest.fixture
def hookeslaw2dplainstress_real():
    return HookesLaw2dPlaneStress(E=205e9, nu=0.3)


def test_hookeslaw2dplainstress_real_init(hookeslaw2dplainstress_real):
    assert hookeslaw2dplainstress_real._E == 205e9
    assert hookeslaw2dplainstress_real._nu == 0.3


def test_hookeslaw2dplainstress_real_properties(hookeslaw2dplainstress_real):
    assert hookeslaw2dplainstress_real.E == hookeslaw2dplainstress_real._E
    assert hookeslaw2dplainstress_real.nu == hookeslaw2dplainstress_real._nu
    np.testing.assert_approx_equal(hookeslaw2dplainstress_real.G, 78846153e3, significant=5)
    np.testing.assert_approx_equal(hookeslaw2dplainstress_real.K, 170838383e3, significant=5)


# (s11, s22, s33, s12), (e11, e22, e33, g12)
data2Dstress_real = np.array([
    [[100e6, 0.0, 0.0, 0.0], [1./2050, -3./20500, -3./20500, 0.0]],
    [[0.0, 100e6, 0.0, 0.0], [-3./20500, 1./2050, -3./20500, 0.0]],
    [[0.0, 0.0, 0.0, 100e6], [0.0, 0.0, 0.0, 13./10250]],
    [[100e6, 200e6, 0.0, 50e6], [1./5125, 17./20500, -9./20500, 13./20500]],
    [[-50e6, 100e6, 0.0, 200e6], [-2./5125, 23./41000, -3./41000, 13./5125]],
    [[50e6, -50e6, 0.0, 100e6], [13./41000, -13./41000, 0.0, 13./10250]]
])


@pytest.mark.parametrize('stress, expected', map(tuple, data2Dstress_real))
def test_hookeslaw2dplainstress_real_strain_scalar(hookeslaw2dplainstress_real, stress, expected):
    stress = dict(s11=stress[0], s22=stress[1], s12=stress[3])
    np.testing.assert_allclose(hookeslaw2dplainstress_real.strain(**stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', map(tuple, data2Dstress_real))
def test_hookeslaw2dplainstress_real_stress_scalar(hookeslaw2dplainstress_real, expected, strain):
    strain = dict(e11=strain[0], e22=strain[1], g12=strain[3])
    np.testing.assert_allclose(hookeslaw2dplainstress_real.stress(**strain), expected[[0, 1, 3]], rtol=1e-5)


@pytest.mark.parametrize('stress, expected', [(data2Dstress_real[:, 0], data2Dstress_real[:, 1])])
def test_hookeslaw2dplainstress_real_strain_array(hookeslaw2dplainstress_real, stress, expected):
    stress = dict(s11=stress[:, 0], s22=stress[:, 1], s12=stress[:, 3])
    np.testing.assert_allclose(np.array(hookeslaw2dplainstress_real.strain(**stress)).T, expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(data2Dstress_real[:, 0], data2Dstress_real[:, 1])])
def test_hookeslaw2dplainstress_real_stress_array(hookeslaw2dplainstress_real, expected, strain):
    strain = dict(e11=strain[:, 0], e22=strain[:, 1], g12=strain[:, 3])
    np.testing.assert_allclose(np.array(hookeslaw2dplainstress_real.stress(**strain)).T,
                               expected[:, (0, 1, 3)], rtol=1e-5)


# HookesLaw2dPlaneStrain
@pytest.fixture
def hookeslaw2dplainstrain():
    return HookesLaw2dPlaneStrain(E=2., nu=0.1)


@pytest.mark.parametrize('nu', [-1.1, 0.6])
def test_hookeslaw2dplainstrain_init(nu):
    with pytest.raises(ValueError,
                       match=re.escape('Possion\'s ratio nu is %.2f but must be -1 <= nu <= 1./2.' % nu)):
        HookesLaw2dPlaneStrain(E=2., nu=nu)


def test_hookeslaw2dplainstrain_wrongshapestress(hookeslaw2dplainstrain):
    with pytest.raises(ValueError,
                       match=re.escape('Components\' shape is not consistent.')):
        hookeslaw2dplainstrain.stress(e11=np.zeros((1, 2)), e22=np.zeros((1, 2)), g12=np.zeros((2, 1)))


def test_hookeslaw2dplainstrain_wrongshapestrain(hookeslaw2dplainstrain):
    with pytest.raises(ValueError,
                       match=re.escape('Components\' shape is not consistent.')):
        hookeslaw2dplainstrain.strain(s11=np.zeros((1, 2)), s22=np.zeros((1, 2)), s12=np.zeros((2, 1)))


def test_hookeslaw2dplainstrain_init(hookeslaw2dplainstrain):
    assert hookeslaw2dplainstrain._E == 2.
    assert hookeslaw2dplainstrain._nu == 0.1


def test_hookeslaw2dplainstrain_properties(hookeslaw2dplainstrain):
    assert hookeslaw2dplainstrain.E == hookeslaw2dplainstrain._E
    assert hookeslaw2dplainstrain.nu == hookeslaw2dplainstrain._nu
    np.testing.assert_approx_equal(hookeslaw2dplainstrain.G, 0.9090909, significant=5)
    np.testing.assert_approx_equal(hookeslaw2dplainstrain.K, 0.8333333, significant=5)


# (s11, s22, s33, s12), (e11, e22, e33, g12)
data2Dstrain = np.array([
    [[0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]],
    [[1.0, 0.0, 0.1, 0.0], [99./200, -11./200, 0.0, 0.0]],
    [[0.0, 1.0, 0.1, 0.0], [-11./200, 99./200, 0.0, 0.0]],
    [[0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 1.1]],
    [[1.0, 2.0, 0.3, 0.5], [77./200, 187./200, 0.0, 0.55]],
    [[-0.5, 1.0, 0.05, 2.0], [-121./400, 209./400, 0.0, 2.2]],
    [[0.5, -0.5, 0.0, 1.0], [11./40, -11./40, 0.0, 1.1]]
])


@pytest.mark.parametrize('stress, expected', map(tuple, data2Dstrain))
def test_hookeslaw2dplainstrain_strain_scalar(hookeslaw2dplainstrain, stress, expected):
    stress = dict(s11=stress[0], s22=stress[1], s12=stress[3])
    np.testing.assert_allclose(hookeslaw2dplainstrain.strain(**stress), expected[[0, 1, 3]], rtol=1e-5)


@pytest.mark.parametrize('expected, strain', map(tuple, data2Dstrain))
def test_hookeslaw2dplainstrain_stress_scalar(hookeslaw2dplainstrain, expected, strain):
    strain = dict(e11=strain[0], e22=strain[1], g12=strain[3])
    np.testing.assert_allclose(hookeslaw2dplainstrain.stress(**strain), expected, rtol=1e-5)


@pytest.mark.parametrize('stress, expected', [(data2Dstrain[:, 0], data2Dstrain[:, 1])])
def test_hookeslaw2dplainstrain_strain_array(hookeslaw2dplainstrain, stress, expected):
    stress = dict(s11=stress[:, 0], s22=stress[:, 1], s12=stress[:, 3])
    np.testing.assert_allclose(np.array(hookeslaw2dplainstrain.strain(**stress)).T, expected[:, (0, 1, 3)], rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(data2Dstrain[:, 0], data2Dstrain[:, 1])])
def test_hookeslaw2dplainstrain_stress_array(hookeslaw2dplainstrain, expected, strain):
    strain = dict(e11=strain[:, 0], e22=strain[:, 1], g12=strain[:, 3])
    np.testing.assert_allclose(np.array(hookeslaw2dplainstrain.stress(**strain)).T, expected, rtol=1e-5)


@pytest.fixture
def hookeslaw2dplainstrain_real():
    return HookesLaw2dPlaneStrain(E=205e9, nu=0.3)


def test_hookeslaw2dplainstrain_real_init(hookeslaw2dplainstrain_real):
    assert hookeslaw2dplainstrain_real._E == 205e9
    assert hookeslaw2dplainstrain_real._nu == 0.3


def test_hookeslaw2dplainstrain_real_properties(hookeslaw2dplainstrain_real):
    assert hookeslaw2dplainstrain_real.E == hookeslaw2dplainstrain_real._E
    assert hookeslaw2dplainstrain_real.nu == hookeslaw2dplainstrain_real._nu
    np.testing.assert_approx_equal(hookeslaw2dplainstrain_real.G, 78846153e3, significant=5)
    np.testing.assert_approx_equal(hookeslaw2dplainstrain_real.K, 170838383e3, significant=5)


# (s11, s22, s33, s12), (e11, e22, e33, g12)
data2Dstrain_real = np.array([
    [[100e6, 0.0, 30.0e6, 0.0], [91./205000, -39./205000, 0.0, 0.0]],
    [[0.0, 100e6, 30.0e6, 0.0], [-39./205000, 91./205000, 0.0, 0.0]],
    [[0.0, 0.0, 0.0, 100e6], [0.0, 0.0, 0.0, 13./10250]],
    [[100e6, 200e6, 90.0e6, 50e6], [13./205000, 143./205000, 0.0, 13./20500]],
    [[-50e6, 100e6, 15.0e6, 200e6], [-169./410000, 221./410000, 0.0, 13./5125]],
    [[50e6, -50e6, 0.0, 100e6], [13./41000, -13./41000, 0.0, 13./10250]]
])


@pytest.mark.parametrize('stress, expected', map(tuple, data2Dstrain_real))
def test_hookeslaw2dplainstrain_real_strain_scalar(hookeslaw2dplainstrain_real, stress, expected):
    stress = dict(s11=stress[0], s22=stress[1], s12=stress[3])
    np.testing.assert_allclose(hookeslaw2dplainstrain_real.strain(**stress), expected[[0, 1, 3]], rtol=1e-5)


@pytest.mark.parametrize('expected, strain', map(tuple, data2Dstrain_real))
def test_hookeslaw2dplainstrain_real_stress_scalar(hookeslaw2dplainstrain_real, expected, strain):
    strain = dict(e11=strain[0], e22=strain[1], g12=strain[3])
    np.testing.assert_allclose(hookeslaw2dplainstrain_real.stress(**strain), expected, rtol=1e-5)


@pytest.mark.parametrize('stress, expected', [(data2Dstrain_real[:, 0], data2Dstrain_real[:, 1])])
def test_hookeslaw2dplainstrain_real_strain_array(hookeslaw2dplainstrain_real, stress, expected):
    stress = dict(s11=stress[:, 0], s22=stress[:, 1], s12=stress[:, 3])
    np.testing.assert_allclose(np.array(hookeslaw2dplainstrain_real.strain(**stress)).T,
                               expected[:, (0, 1, 3)], rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(data2Dstrain_real[:, 0], data2Dstrain_real[:, 1])])
def test_hookeslaw2dplainstrain_real_stress_array(hookeslaw2dplainstrain_real, expected, strain):
    strain = dict(e11=strain[:, 0], e22=strain[:, 1], g12=strain[:, 3])
    np.testing.assert_allclose(np.array(hookeslaw2dplainstrain_real.stress(**strain)).T, expected, rtol=1e-5)


# HookesLaw3d
@pytest.fixture
def hookeslaw3d():
    return HookesLaw3d(E=2., nu=0.1)


@pytest.mark.parametrize('nu', [-1.1, 0.6])
def test_hookeslaw3d_init(nu):
    with pytest.raises(ValueError,
                       match=re.escape('Possion\'s ratio nu is %.2f but must be -1 <= nu <= 1./2.' % nu)):
        HookesLaw2dPlaneStrain(E=2., nu=nu)


def test_hookeslaw3d_wrongshapestress(hookeslaw3d):
    with pytest.raises(ValueError,
                       match=re.escape('Components\' shape is not consistent.')):
        hookeslaw3d.stress(
            e11=np.zeros((1, 2)), e22=np.zeros((1, 2)), e33=np.zeros((2, 1)),
            g12=np.zeros((1, 2)), g13=np.zeros((1, 2)), g23=np.zeros((1, 1)))


def test_hookeslaw3d_wrongshapestrain(hookeslaw3d):
    with pytest.raises(ValueError,
                       match=re.escape('Components\' shape is not consistent.')):
        hookeslaw3d.strain(
            s11=np.zeros((1, 2)), s22=np.zeros((1, 2)), s33=np.zeros((2, 1)),
            s12=np.zeros((1, 2)), s23=np.zeros((1, 1)), s13=np.zeros((2, 1)))


def test_hookeslaw3d_init(hookeslaw3d):
    assert hookeslaw3d._E == 2.
    assert hookeslaw3d._nu == 0.1


def test_hookeslaw3d_properties(hookeslaw3d):
    assert hookeslaw3d.E == hookeslaw3d._E
    assert hookeslaw3d.nu == hookeslaw3d._nu
    np.testing.assert_approx_equal(hookeslaw3d.G, 0.9090909, significant=5)
    np.testing.assert_approx_equal(hookeslaw3d.K, 0.8333333, significant=5)


# (s11, s22, s33, s12, s13, s23), (e11, e22, e33, g12, g13, g23)
data3D = np.array([
    [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]],
    [[1.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.5, -0.05, -0.05, 0.0, 0.0, 0.0]],
    [[0.0, 1.0, 0.0, 0.0, 0.0, 0.0], [-0.05, 0.5, -0.05, 0.0, 0.0, 0.0]],
    [[0.0, 0.0, 1.0, 0.0, 0.0, 0.0], [-0.05, -0.05, 0.5, 0.0, 0.0, 0.0]],
    [[0.0, 0.0, 0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.1, 0.0, 0.0]],
    [[0.0, 0.0, 0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0, 1.1, 0.0]],
    [[0.0, 0.0, 0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 0.0, 0.0, 1.1]],
    [[1.0, 2.0, 3.0, 0.5, 0.5, 0.5], [0.25, 0.8, 1.35, 0.55, 0.55, 0.55]],
    [[-3.0, 2.0, 1.0, 1., 0.5, 2.0], [-1.65, 1.1, 0.55, 1.1, 0.55, 2.2]],
    [[-1.0, 3.0, 0.5, 2.0, -1.0, 2.0], [-0.675, 1.525, 0.15, 2.2, -1.1, 2.2]],
    [[0.5, -0.5, 0.0, 3.0, 2.0, 1.0], [0.275, -0.275, 0.0, 3.3, 2.2, 1.1]]
])


@pytest.mark.parametrize('stress, expected', map(tuple, data3D))
def test_hookeslaw3d_strain_scalar(hookeslaw3d, stress, expected):
    stress = dict(s11=stress[0], s22=stress[1], s33=stress[2], s12=stress[3], s13=stress[4], s23=stress[5])
    np.testing.assert_allclose(hookeslaw3d.strain(**stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', map(tuple, data3D))
def test_hookeslaw3d_stress_scalar(hookeslaw3d, expected, strain):
    strain = dict(e11=strain[0], e22=strain[1], e33=strain[2], g12=strain[3], g13=strain[4], g23=strain[5])
    np.testing.assert_allclose(hookeslaw3d.stress(**strain), expected, rtol=1e-5)


@pytest.mark.parametrize('stress, expected', [(data3D[:, 0], data3D[:, 1])])
def test_hookeslaw3d_strain_array(hookeslaw3d, stress, expected):
    stress = dict(s11=stress[:, 0], s22=stress[:, 1], s33=stress[:, 2],
                  s12=stress[:, 3], s13=stress[:, 4], s23=stress[:, 5])
    np.testing.assert_allclose(np.array(hookeslaw3d.strain(**stress)).T, expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(data3D[:, 0], data3D[:, 1])])
def test_hookeslaw3d_stress_array(hookeslaw3d, expected, strain):
    strain = dict(e11=strain[:, 0], e22=strain[:, 1], e33=strain[:, 2],
                  g12=strain[:, 3], g13=strain[:, 4], g23=strain[:, 5])
    np.testing.assert_allclose(np.array(hookeslaw3d.stress(**strain)).T, expected, rtol=1e-5)


@pytest.fixture
def hookeslaw3d_real():
    return HookesLaw3d(E=205e9, nu=0.3)


def test_hookeslaw3d_real_init(hookeslaw3d_real):
    assert hookeslaw3d_real._E == 205e9
    assert hookeslaw3d_real._nu == 0.3


def test_hookeslaw3d_real_properties(hookeslaw3d_real):
    assert hookeslaw3d_real.E == hookeslaw3d_real._E
    assert hookeslaw3d_real.nu == hookeslaw3d_real._nu
    np.testing.assert_approx_equal(hookeslaw3d_real.G, 78846153e3, significant=5)
    np.testing.assert_approx_equal(hookeslaw3d_real.K, 170838383e3, significant=5)


# (s11, s22, s33, s12, s13, s23), (e11, e22, e33, g12, g13, g23)
data3D_real = np.array([
    [[100e6, 0.0, 0.0, 0.0, 0.0, 0.0], [1./2050, -3./20500, -3./20500, 0.0, 0.0, 0.0]],
    [[0.0, 100e6, 0.0, 0.0, 0.0, 0.0], [-3./20500, 1./2050, -3./20500, 0.0, 0.0, 0.0]],
    [[0.0, 0.0, 100e6, 0.0, 0.0, 0.0], [-3./20500, -3./20500, 1./2050, 0.0, 0.0, 0.0]],
    [[0.0, 0.0, 0.0, 100e6, 0.0, 0.0], [0.0, 0.0, 0.0, 13./10250, 0.0, 0.0]],
    [[0.0, 0.0, 0.0, 0.0, 100e6, 0.0], [0.0, 0.0, 0.0, 0.0, 13./10250, 0.0]],
    [[0.0, 0.0, 0.0, 0.0, 0.0, 100e6], [0.0, 0.0, 0.0, 0.0, 0.0, 13./10250]],
    [[100e6, 100e6, 100e6, 100e6, 100e6, 100e6], [1./5125, 1./5125, 1./5125, 13./10250, 13./10250, 13./10250]],
    [[-300e6, 200e6, 100e6, 100e6, 50e6, 200e6], [-39./20500, 13./10250, 13./20500, 13./10250, 13./20500, 13./5125]],
    [[-100e6, 300e6, 50e6, 200e6, -100e6, 200e6], [-1./1000, 63./41000, -1./20500, 13./5125, -13./10250, 13./5125]],
    [[50e6, -50e6, 0.0, 300e6, 200e6, 100e6], [13./41000, -13./41000, 0.0, 39./10250, 13./5125, 13./10250]]
])


@pytest.mark.parametrize('stress, expected', map(tuple, data3D_real))
def test_hookeslaw3d_real_strain_scalar(hookeslaw3d_real, stress, expected):
    stress = dict(s11=stress[0], s22=stress[1], s33=stress[2], s12=stress[3], s13=stress[4], s23=stress[5])
    np.testing.assert_allclose(hookeslaw3d_real.strain(**stress), expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', map(tuple, data3D_real))
def test_hookeslaw3d_real_stress_scalar(hookeslaw3d_real, expected, strain):
    strain = dict(e11=strain[0], e22=strain[1], e33=strain[2], g12=strain[3], g13=strain[4], g23=strain[5])
    np.testing.assert_allclose(hookeslaw3d_real.stress(**strain), expected, rtol=1e-5)


@pytest.mark.parametrize('stress, expected', [(data3D_real[:, 0], data3D_real[:, 1])])
def test_hookeslaw3d_real_strain_array(hookeslaw3d_real, stress, expected):
    stress = dict(s11=stress[:, 0], s22=stress[:, 1], s33=stress[:, 2],
                  s12=stress[:, 3], s13=stress[:, 4], s23=stress[:, 5])
    np.testing.assert_allclose(np.array(hookeslaw3d_real.strain(**stress)).T, expected, rtol=1e-5)


@pytest.mark.parametrize('expected, strain', [(data3D_real[:, 0], data3D_real[:, 1])])
def test_hookeslaw3d_real_stress_array(hookeslaw3d_real, expected, strain):
    strain = dict(e11=strain[:, 0], e22=strain[:, 1], e33=strain[:, 2],
                  g12=strain[:, 3], g13=strain[:, 4], g23=strain[:, 5])
    np.testing.assert_allclose(np.array(hookeslaw3d_real.stress(**strain)).T, expected, rtol=1e-5)



# Copyright (c) 2019-2023 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pylife.strength.solidity as SLD

import pandas as pd
import numpy as np
import pytest

@pytest.fixture
def homogenious_collective():
    return pd.Series(2342.4711,
                     name='cycles',
                     index=pd.IntervalIndex.from_tuples([(1002., 1003.)], name='range'))

@pytest.fixture
def real_life_collective():
    S = np.array([
        0.015625, 0.046875, 0.078125, 0.109375, 0.140625, 0.171875,
        0.203125, 0.234375, 0.265625, 0.296875, 0.328125, 0.359375,
        0.390625, 0.421875, 0.453125, 0.484375, 0.515625, 0.546875,
        0.578125, 0.609375, 0.640625, 0.671875, 0.703125, 0.734375,
        0.765625, 0.796875, 0.828125, 0.859375, 0.890625, 0.921875,
        0.953125, 0.984375, 1.015625, 1.046875, 1.078125, 1.109375,
        1.140625, 1.171875, 1.203125, 1.234375, 1.265625, 1.296875,
        1.328125, 1.359375, 1.390625, 1.421875, 1.453125, 1.484375,
        1.515625, 1.546875, 1.578125, 1.609375, 1.640625, 1.671875,
        1.703125, 1.734375, 1.765625, 1.796875, 1.828125, 1.859375,
        1.890625, 1.921875, 1.953125, 1.984375])

    mids = (S[:-1] + S[1:]) / 2.
    start = 2. * S[0] - mids[0]
    end = 2 * S[-1] - mids[-1]
    amplitudes = np.concatenate([[start], mids, [end]])

    idx = pd.IntervalIndex.from_breaks(amplitudes*2., name='range')

    N = np.array([
        190733.0, 190733.0, 181542.0, 169283.0, 153244.0, 124473.0,
        113752.0, 86802.0, 81117.0, 57718.0, 55726.0, 41750.0, 39033.0,
        30847.0, 28007.0, 23177.0, 20532.0, 17860.0, 15454.0, 13867.0,
        11601.0, 10765.0, 8831.0, 8449.0, 6682.0, 6504.0, 5044.0, 4967.0,
        3801.0, 3720.0, 2775.0, 2671.0, 2014.0, 1825.0, 1403.0, 1195.0,
        954.0, 717.0, 640.0, 418.0, 393.0, 257.0, 256.0, 152.0, 152.0,
        104.0, 104.0, 63.0, 62.0, 34.0, 32.0, 18.0, 18.0, 12.0, 9.0, 5.0,
        3.0, 3.0, 1.0, 1.0, 0, 0, 0, 0]
    )

    decumulated = np.append(-np.diff(N), 0.0)

    return pd.Series(decumulated, index=idx, name='cycles')


def test_solidity_haibach_homogenious(homogenious_collective):
    np.testing.assert_almost_equal(homogenious_collective.solidity.haibach(k=6), 1.0)


def test_solidity_fkm_homogenious(homogenious_collective):
    np.testing.assert_almost_equal(homogenious_collective.solidity.fkm(k=6), 1.0)


def test_solidity_haibach(real_life_collective):
    np.testing.assert_almost_equal(real_life_collective.solidity.haibach(k=6), 0.00115875)


def test_solidity_fkm(real_life_collective):
    np.testing.assert_almost_equal(real_life_collective.solidity.fkm(k=6), 0.32408968)
